@@ -18,6 +18,7 @@ void
 generate_old_dump(void)
 {
 	int			dbnum;
+	mode_t		old_umask;
 
 	prep_status("Creating dump of global objects");
 
@@ -32,6 +33,13 @@ generate_old_dump(void)
 
 	prep_status("Creating dump of database schemas\n");
 
+	/*
+	 * Set umask for this function, all functions it calls, and all
+	 * subprocesses/threads it creates.  We can't use fopen_priv()
+	 * as Windows uses threads and umask is process-global.
+	 */
+	old_umask = umask(S_IRWXG | S_IRWXO);
+
 	/* create per-db dump files */
 	for (dbnum = 0; dbnum < old_cluster.dbarr.ndbs; dbnum++)
 	{
@@ -66,6 +74,8 @@ generate_old_dump(void)
 	while (reap_child(true) == true)
 		;
 
+	umask(old_umask);
+
 	end_progress_output();
 	check_ok();
 }
