@@ -2,9 +2,9 @@
  * Portions Copyright (c) 1987, 1993, 1994
  * The Regents of the University of California.  All rights reserved.
  *
- * Portions Copyright (c) 2003, PostgreSQL Global Development Group
+ * Portions Copyright (c) 2003-2013, PostgreSQL Global Development Group
  *
- * $Header: /cvsroot/pgsql/src/include/getopt_long.h,v 1.4 2003/08/08 21:42:31 momjian Exp $
+ * src/include/getopt_long.h
  */
 #ifndef GETOPT_LONG_H
 #define GETOPT_LONG_H
@@ -19,9 +19,6 @@ extern int	optind;
 extern int	optopt;
 extern char *optarg;
 
-/* Some systems have this, otherwise you need to define it somewhere. */
-extern int	optreset;
-
 #ifndef HAVE_STRUCT_OPTION
 
 struct option
